///////////////////////////////////////////////////////////////////////////////
// PjTomography class
//
// Last update: 23/01/2008
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Native JAVA imports
///////////////////////////////////////////////////////////////////////////////
import java.awt.Color;
import java.io.BufferedReader;
import java.applet.AudioClip;

import jv.geom.PgElementSet;
import jv.geom.PgPolygonSet;
import jv.loader.PgJvxLoader;
import jv.number.PuDouble;
import jv.object.PsDebug;
import jv.object.PsUtil;
import jv.project.PgJvxSrc;
import jv.project.PjProject;
import jv.project.PvCameraIf;
import jv.project.PvDisplayIf;
import jv.vecmath.PdVector;
import jvx.geom.PwCleanMesh;
import jvx.geom.PwIntersectWithFunction;
import jvx.geom.PwTransform;
import jvx.loader.PgOffLoader;



///////////////////////////////////////////////////////////////////////////////
// Tomography applet
///////////////////////////////////////////////////////////////////////////////
public class PjTomography extends PjProject 
{
	protected int idiom;
	protected String codebase;
	protected double initial_transparency;

	protected PuDouble transparency;
	protected PuDouble perspective;

	protected int numberOfVertices;
	protected int numberOfEdges;
	protected int numberOfElements;

	protected PgElementSet poly;
	protected PgElementSet copy;

	protected PvDisplayIf disp;
    protected PvCameraIf cam;

	protected PgElementSet planeXY;
	protected PgElementSet planeXZ;
	protected PgElementSet planeYZ;
	protected double xMin;
	protected double xMax;
	protected double yMin;
	protected double yMax;
	protected double zMin;
	protected double zMax;
	protected double delta;
	protected double xO;
	protected double yO;
	protected double zO;    	
	
	protected PuDouble translation_factor;	
	protected PgElementSet cutplane;	
	
	protected PwIntersectWithFunction m_pwIntersectWithFunction;
	protected PgPolygonSet m_slice;
	
	protected boolean firstTime = true;

	protected AudioClip rightAnswerSoundData;
	protected AudioClip wrongAnswerSoundData;
	
	protected boolean showHelp = false;
	protected boolean showSolution = false;
	
	protected int number_of_help_slices = 11;
	protected PgPolygonSet[] help_slices;	
	

	//////////////////////////////////////////////////////////////////////////
	// Default constructor
	//////////////////////////////////////////////////////////////////////////
	public PjTomography(String input_codebase, 
                        AudioClip input_rightAnswerSoundData,
                        AudioClip input_wrongAnswerSoundData,			
			            int input_idiom,  
			            double input_initial_transparency) 
	{
		super("Spatial Geometry Applet");

		codebase = input_codebase;
		rightAnswerSoundData = input_rightAnswerSoundData;
	    wrongAnswerSoundData = input_wrongAnswerSoundData;
	    
		idiom = input_idiom;
		initial_transparency = input_initial_transparency;

		if (getClass() == PjTomography.class) 
		{
			init();
		};
	} // End of PjTomography()

	// ////////////////////////////////////////////////////////////////////////
	// Initialize variables
	// ////////////////////////////////////////////////////////////////////////
	public void init() 
	{
		poly = new PgElementSet(3);

		planeXY = new PgElementSet(3);
		planeXY.setNumElements(1);
		planeXY.setNumVertices(4);
		planeXZ = new PgElementSet(3);
		planeXZ.setNumElements(1);
		planeXZ.setNumVertices(4);
		planeYZ = new PgElementSet(3);
		planeYZ.setNumElements(1);
		planeYZ.setNumVertices(4);

		copy = new PgElementSet(3);

		// Transparency slider
		transparency = new PuDouble("Transparncia:", this);
		transparency.setDefBounds(0.0, 1.0, 0.01, 0.05);
		transparency.setDefValue(initial_transparency);
		transparency.init();
		transparency.setEnabled(true);
		
		// Perspective slider
		perspective = new PuDouble("Perspectiva:", this);
		perspective.setDefBounds(0.0, 1.5, 0.01, 0.05);
		perspective.setDefValue(0.5);
		perspective.init();
		perspective.setEnabled(true);		
		
	    // Rotation around e-axis slider configuration
	    translation_factor = new PuDouble("Translao:", this);
	    translation_factor.setDefBounds(-1, 1, 0.1, 0.2);
	    translation_factor.setDefValue(0.0);
	    translation_factor.init();	  
	    
        // Cut plane
        cutplane = new PgElementSet(3); // This is a 3D object
        cutplane.setNumElements(1);
        cutplane.setNumVertices(4); // Our poly has 8 vertices
        cutplane.setVertex(0,  1,  1, 0);
        cutplane.setVertex(1, -1,  1, 0);
        cutplane.setVertex(2, -1, -1, 0);
        cutplane.setVertex(3,  1, -1, 0);
        cutplane.setElement(0, 0, 1, 2, 3);
        cutplane.showElementColors(true);
        cutplane.showElements(true);
        cutplane.showEdges(true);
        cutplane.setGlobalElementColor(Color.BLUE);
        cutplane.setTransparency(0.7);
        cutplane.showTransparency(true);
        
        // Slice
        m_slice = new PgPolygonSet(3);
        m_slice.setName("Slice");
        
        // Help slices
        help_slices = new PgPolygonSet[number_of_help_slices];
        for (int i = 0; i < number_of_help_slices; i++)
        {
            help_slices[i] = new PgPolygonSet(3);
            help_slices[i].setName("Help Slice " + i);        	
        }
        
        m_pwIntersectWithFunction = new PwIntersectWithFunction();

	}; // End of init()

	// ////////////////////////////////////////////////////////////////////////
	// Load geometry file
	// ////////////////////////////////////////////////////////////////////////
	public void loadGeometryFromFile(GameObject object) 
	{		
		if (showSolution == true)
		{
			removeGeometry(poly);
			showSolution = false;
			removeHelpSlices();
		}
		
		if (showHelp == true)
		{
			for (int i = 0; i < number_of_help_slices; i++)
			{
				removeGeometry(help_slices[i]);
			};
			showHelp = false;
		}
		
		String filename = object.getFileName();
		String filetype = object.getFileType();
		String pathname = codebase + filename + "." + filetype;

		/*
			URL fullpath = new URL(pathname);
			java.io.InputStream r = fullpath.openStream();
			BufferedReader in = new BufferedReader(new InputStreamReader(r));
			*/
			BufferedReader in = PsUtil.open(pathname);
			if (in == null) {
				PsDebug.warning("could not open file  " + pathname + "...");
				System.out.println("could not open file  " + pathname + "...");
				return;
			} 

				if (firstTime == false)
				{
					m_pwIntersectWithFunction.cancel();
					copy.copy(poly);
				}
				// Loading polyhedron
				PgJvxSrc[] geometries;
				if (filetype.equals("off")) 
				{
					PgOffLoader offloader = new PgOffLoader();
					offloader.load(in);
					geometries = offloader.getGeometries();
					poly.setJvx(geometries[0]);
				}
				if (filetype.equals("jvx")) 
				{
					PgJvxLoader jvxloader = new PgJvxLoader();
					jvxloader.load(in);
					geometries = jvxloader.getGeometries();
					poly.setJvx(geometries[0]);
				}

				// Transform polyhedron
				if (object.getApplyTransform() == true)
				{
					double angle1 = object.getRot1() * Math.PI / 180.0;
					double angle2 = object.getRot2() * Math.PI / 180.0;
					double angle3 = object.getRot3() * Math.PI / 180.0;

					PwTransform.rotate(poly, angle1, 0);
					PwTransform.rotate(poly, angle2, 1);
					PwTransform.rotate(poly, angle3, 2);

					poly.update(poly);										
				}
				
				// Configuring the polyhedron
				poly.setTransparency(initial_transparency);
				if (transparency.getValue() > 0.05) 
				{
					poly.showTransparency(true);
				} 
				else 
				{
					poly.showTransparency(false);
				}
				poly.showElementColors(true);
				poly.showVertices(object.getShowVertices());				
				poly.showEdges(object.getShowEdges());
				poly.showElements(object.getShowElements());				
				poly.showVertexOutline(true);
				poly.setGlobalEdgeColor(Color.BLACK);
				poly.setGlobalVertexColor(Color.WHITE);
				poly.showVertexColors(true);

				PwCleanMesh.identifyVertices(poly, 1E-010D);
				poly.makeNeighbour();
				poly.close();
				
				xMin = (poly.getBounds())[0].getEntry(0);
				yMin = (poly.getBounds())[0].getEntry(1);
				zMin = (poly.getBounds())[0].getEntry(2);
				xMax = (poly.getBounds())[1].getEntry(0);
				yMax = (poly.getBounds())[1].getEntry(1);
				zMax = (poly.getBounds())[1].getEntry(2);

				double deltaX = xMax - xMin;
				double deltaY = yMax - yMin;
				double deltaZ = zMax - zMin;

				delta = deltaZ;
				if (delta < deltaX) 
				{
					delta = deltaX;
				}
				if (delta < deltaY) 
				{
					delta = deltaY;
				}
				delta = 1.1 * Math.sqrt(3.0) * delta;

				xO = xMin - (delta - deltaX) / 2.0;
				yO = yMin - (delta - deltaY) / 2.0;
				zO = zMin - (delta - deltaZ) / 2.0;

				// Plane XY
				planeXY.setVertex(0, xO, yO, zO);
				planeXY.setVertex(1, xO, yO + delta, zO);
				planeXY.setVertex(2, xO + delta, yO + delta, zO);
				planeXY.setVertex(3, xO + delta, yO, zO);
				planeXY.setElement(0, 0, 1, 2, 3);
				planeXY.showElementColors(true);
				planeXY.showElements(true);
				planeXY.showEdges(true);
				planeXY.setGlobalElementColor(Color.LIGHT_GRAY);
				planeXY.setTransparency(0.7);
				planeXY.showTransparency(true);

				// Plane XZ
				planeXZ.setVertex(0, xO, yO, zO);
				planeXZ.setVertex(1, xO + delta, yO, zO);
				planeXZ.setVertex(2, xO + delta, yO, zO + delta);
				planeXZ.setVertex(3, xO, yO, zO + delta);
				planeXZ.setElement(0, 0, 1, 2, 3);
				planeXZ.showElementColors(true);
				planeXZ.showElements(true);
				planeXZ.showEdges(true);
				planeXZ.setGlobalElementColor(Color.LIGHT_GRAY);
				planeXZ.setTransparency(0.7);
				planeXZ.showTransparency(true);

				// Plane YZ
				planeYZ.setVertex(0, xO, yO, zO);
				planeYZ.setVertex(1, xO, yO + delta, zO);
				planeYZ.setVertex(2, xO, yO + delta, zO + delta);
				planeYZ.setVertex(3, xO, yO, zO + delta);
				planeYZ.setElement(0, 0, 1, 2, 3);
				planeYZ.showElementColors(true);
				planeYZ.showElements(true);
				planeYZ.showEdges(true);
				planeYZ.setGlobalElementColor(Color.LIGHT_GRAY);
				planeYZ.setTransparency(0.7);
				planeYZ.showTransparency(true);

				poly.update(poly); // Don't forget to update!

				planeXY.update(planeXY);
				planeXZ.update(planeXZ);
				planeYZ.update(planeYZ);

				PwCleanMesh.identifyVertices(poly, 1E-010D);
				poly.makeNeighbour();

				poly.setTransparency(initial_transparency);
				poly.showTransparency(true);
				poly.update(poly);
				
				PwCleanMesh.identifyVertices(copy, 1E-010D);
				copy.makeNeighbour();

				copy.setTransparency(initial_transparency);
				copy.showTransparency(true);
				copy.update(copy);				
								
				// Copy of poly
				copy.copy(poly);

				// Translation factor
			    translation_factor.setBounds(zO + delta/100.0, zO + delta - delta/100.0, 0.1, 0.2);
			    translation_factor.setValue(zO + delta/2.0);

			    m_pwIntersectWithFunction.setGeometry(copy);			   
		  	    m_pwIntersectWithFunction.setFunction("w");			    

		  	    if (firstTime == false)
		  	    {
		  	    	disp.fit();
		  	    };
		  	    
			    firstTime = false;
			    
			    updateCutPlane();
			    

	} // End of loadGeometryFromFile()


	// ////////////////////////////////////////////////////////////////////////
	// Things to do when the class is started
	// ////////////////////////////////////////////////////////////////////////
	public void start() 
	{
		super.start();

		// Necessary to apply PwWorkshop operations
		PwCleanMesh.identifyVertices(poly, 1E-010D);
		poly.makeNeighbour();

		poly.setTransparency(initial_transparency);
		poly.showTransparency(true);
		// addGeometry(poly);
		addGeometry(planeXY);
		addGeometry(planeXZ);
		addGeometry(planeYZ);
		addGeometry(cutplane);
		addGeometry(m_slice);
		selectGeometry(poly);

		update(this);

		disp = getDisplay(); // The PwUnfold needs a reference to the
		                     // display in order to work properly.

		PdVector d1 = new PdVector(1, 1, 1);
		PdVector d2 = new PdVector(3);
		d2.copy(d1);
		d2.multScalar(-1);
		cam = disp.getCamera();
		cam.setPosition(d1);
		cam.setViewDir(d2);
		cam.setRoll(0);
		cam.update(cam);
		disp.fit();		
	} // End of start()

	// ///////////////////////////////////////////////////////////////////////
	// Update the class whenever a child has changed. Method is usually
	// invoked from the children.
	// ///////////////////////////////////////////////////////////////////////
	public boolean update(Object event) 
	{
		if (event == null)
		{
			return (true);
		};

		// Solid transparency
		if (event == transparency) 
		{
			poly.setTransparency(transparency.getValue());
			copy.setTransparency(transparency.getValue());

			if (transparency.getValue() > 0.05)
			{
				poly.showTransparency(true);
				copy.showTransparency(true);
			} 
			else 
			{
				poly.showTransparency(false);
				copy.showTransparency(false);
			}

			poly.update(poly);

			return (true);
		};
		
		// Perspective
		if (event == perspective) 
		{
            cam.setFieldOfView(perspective.getValue());
            cam.update(cam);
			return (true);
		};
		

/*		
		// Rotation
		if (event == rot1 || event == rot2 || event == rot3) 
		{
			double angle1 = rot1.getValue() * Math.PI / 180.0;
			double angle2 = rot2.getValue() * Math.PI / 180.0;
			double angle3 = rot3.getValue() * Math.PI / 180.0;

			poly.copy(original);

			PwTransform.rotate(poly, angle1, 0);
			PwTransform.rotate(poly, angle2, 1);
			PwTransform.rotate(poly, angle3, 2);

			poly.update(poly);
			
			copy.copy(poly);
			
			m_pwIntersectWithFunction.setGeometry(copy);			   
	  	    m_pwIntersectWithFunction.setFunction("w");				
			
			updateCutPlane();

			return (true);
		}
 */		
		
		// Translation	     
	    if (event == translation_factor)
	    {
	       updateCutPlane();
	       return(true);
	    }		

		return (super.update(event));
	} // End of update()
	
    //////////////////////////////////////////////////////////////////////////
    // updateCutPlane()
    //////////////////////////////////////////////////////////////////////////
    public synchronized void updateCutPlane()
    {
       // Update cut plane
  	   double h = translation_factor.getValue(); 
       cutplane.setVertex(0, xO + delta,  yO + delta, h);
       cutplane.setVertex(1, xO,  yO + delta, h);
       cutplane.setVertex(2, xO, yO, h);
       cutplane.setVertex(3, xO + delta, yO, h);
       cutplane.update(cutplane);
      
	   m_pwIntersectWithFunction.setLevel(h);			    
	   m_pwIntersectWithFunction.intersectWithFunction();

	   PgPolygonSet slice = m_pwIntersectWithFunction.getIntersection();
	   m_slice.copy(slice);
       m_slice.update(m_slice);

       firstTime = false;       
    } // End of updateCutPlane()
    
    //////////////////////////////////////////////////////////////////////////
    // addHelpSlices()
    //////////////////////////////////////////////////////////////////////////
    public synchronized void addHelpSlices()
    {      
       double vMin = zMin + 0.01*(zMax - zMin);
       double vMax = zMax - 0.01*(zMax - zMin);
           	
       for (int i = 0; i < number_of_help_slices; i++)
       {
          double h = vMin + i*(vMax - vMin)/(number_of_help_slices - 1);
   	      m_pwIntersectWithFunction.setLevel(h);			    
	      m_pwIntersectWithFunction.intersectWithFunction();

	      PgPolygonSet slice = m_pwIntersectWithFunction.getIntersection();
	      help_slices[i].copy(slice);
	      help_slices[i].setGlobalPolygonColor(Color.GREEN);
          addGeometry(help_slices[i]);          
       }
       
       showHelp = true;
    } // End of addHelpSices()
    
    //////////////////////////////////////////////////////////////////////////
    // removeHelpSlices()
    //////////////////////////////////////////////////////////////////////////
    public synchronized void removeHelpSlices()
    {
       for (int i = 0; i < number_of_help_slices; i++)
       {
          removeGeometry(help_slices[i]);          
       }
       
       showHelp = false;
    } // End of removeHelpSices()	        
} // End of PjTomography classs
